// watchdog.c V2

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <string.h>

uint16_t WdtTrapAddress __attribute__ ((section (".noinit")));      // store watchdog interrupt return address

static uint8_t isrStackOffset;

void initWatchdog(uint8_t timeout) {
    uint8_t sreg = SREG;
    cli();

    if (MCUSR & (1 << PORF))                             // on power on reset
        WdtTrapAddress = 0;
    MCUSR = 0;

    wdt_enable(timeout);
    WDTCSR |= (1 << WDIE);                              // enable WDT interrupt

    SREG = sreg;
}

void idle(void) {
    uint8_t timeout = (WDTCSR & ((1 << WDP2) | (1 << WDP1) | (1 << WDP0))) | ((WDTCSR & (1 << WDP3) ? 8 : 0));

	wdt_reset();
    WDTCSR |= (1 << WDIE);                              // enable WDT interrupt

    sleep_cpu();

	wdt_reset();
    wdt_enable(timeout);                                // remember timeout
    WDTCSR |= (1 << WDIE);                              // enable WDT interrupt
}

ISR(WDT_vect) {
    if ((_SLEEP_CONTROL_REG & (uint8_t)_SLEEP_ENABLE_MASK)) {
        // watchdog interrupt in idle
        wdt_disable();                                  // prevent reset
    } else {
        // error condition
        // store ISR return address in .noinit section, so it will overcome the reset
        WdtTrapAddress = 2 * (uint16_t) __builtin_return_address(0);       // use gcc extension to retreve ISR return address


        #ifdef ERROR_ShutdownHook
            ERROR_ShutdownHook;
        #endif // ERROR_ShutdownHook

        // *** RESET MCU NOW ***
        while (1) ;                                     // wait for 2nd watchdog overflow
    }
}
